__version__ = "0.1.1"

from quantalpy.binary_qubit_gate import *
from quantalpy.circuit import *
from quantalpy.ctx_funcs import *
from quantalpy.measure import *
from quantalpy.qft import *
from quantalpy.qpu import *
from quantalpy.typing import *
from quantalpy.unary_qubit_gate import *
from quantalpy.utils import *



_ctx_stack = []


class ContextMixin:
    def __enter__(self) -> "ContextMixin":
        _ctx_stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _ctx_stack.pop(-1)


def _get_current_ctx() -> ContextMixin:
    if not _ctx_stack:
        raise RuntimeError("Nothing in context")
    return _ctx_stack[-1]



import numpy as np

from quantalpy.ctx import ContextMixin
from quantalpy.utils import export

PROJECTORS = [
    np.array([[1, 0], [0, 0]], dtype=complex),
    np.array([[0, 0], [0, 1]], dtype=complex),
]


@export
class QPU(ContextMixin):
    def __init__(self, n_qubits: int) -> None:
        self._n_qubits = n_qubits
        self._amplitudes = None
        self.reset()

    @property
    def n_qubits(self) -> int:
        return self._n_qubits

    def reset(self) -> None:
        self._amplitudes = np.zeros(shape=tuple(2 for _ in range(self.n_qubits)))
        self._amplitudes[tuple(0 for _ in range(self.n_qubits))] = 1

    @staticmethod
    def unary_qubit_operation(
        state: np.ndarray[complex], index: int, matrix: np.ndarray[(2, 2), complex]
    ) -> np.ndarray[complex]:
        aux = np.tensordot(matrix, state, axes=(1, index))
        return np.moveaxis(aux, 0, index)

    def apply_unary_qubit_operator(
        self, index: int, matrix: np.ndarray[(2, 2), complex]
    ) -> None:
        self._amplitudes = self.unary_qubit_operation(
            state=self._amplitudes, matrix=matrix, index=index
        )

    @staticmethod
    def binary_qubit_operator(
        state: np.ndarray[complex],
        indices: tuple[int, int],
        matrix: np.ndarray[(4, 4), complex],
    ) -> np.ndarray[complex]:
        matrix = matrix.reshape((2, 2, 2, 2))
        aux = np.tensordot(matrix, state, axes=((2, 3), indices))
        return np.moveaxis(aux, (0, 1), indices)

    def apply_binary_qubit_operator(
        self, indices: tuple[int, int], matrix: np.ndarray[(4, 4), complex]
    ) -> None:
        self._amplitudes = self.binary_qubit_operator(
            state=self._amplitudes, matrix=matrix, indices=indices
        )

    def measure(self, index: int) -> int:
        result_zero_state = self.unary_qubit_operation(
            matrix=PROJECTORS[0], state=self._amplitudes, index=index
        )
        prop_zero = np.linalg.norm(result_zero_state.flatten()) ** 2
        outcome = np.random.choice([0, 1], p=[prop_zero, 1 - prop_zero])
        match outcome:
            case 0:
                self._amplitudes = result_zero_state / np.linalg.norm(
                    result_zero_state.flatten()
                )
            case 1:
                result_one_state = self.unary_qubit_operation(
                    matrix=PROJECTORS[1], state=self._amplitudes, index=index
                )
                self._amplitudes = result_one_state / np.linalg.norm(
                    result_one_state.flatten()
                )
        return outcome



import numpy as np
import numpy.typing as npt

from quantalpy.runnable import Runnable
from quantalpy.unary_qubit_gate import UnaryQubitGate
from quantalpy.qpu import QPU
from quantalpy.utils import export


@export
class BinaryQubitGate(Runnable):
    def __init__(self, matrix: npt.ArrayLike, indices: tuple[int, int]) -> None:
        self.matrix = np.asarray(matrix, dtype=complex)
        self.indices = indices

    def run(self, qpu: QPU) -> None:
        qpu.apply_binary_qubit_operator(indices=self.indices, matrix=self.matrix)


@export
class ControlledQubitGate(BinaryQubitGate):
    def __init__(self, gate: UnaryQubitGate, control: int) -> None:
        matrix = np.block(
            [[np.eye(2), np.zeros((2, 2))], [np.zeros((2, 2)), gate.matrix]]
        )
        super().__init__(matrix=matrix, indices=(control, gate.index))



import functools
import typing as t

from quantalpy.ctx import _get_current_ctx
from quantalpy.qpu import QPU
from quantalpy.runnable import Runnable
from quantalpy.circuit import Circuit
from quantalpy.unary_qubit_gate import Hadamard, PhaseShift
from quantalpy.measure import Measure
from quantalpy.utils import export


def _create_ctx_function(f: t.Type, name: str) -> t.Callable:
    @functools.wraps(f, assigned=("__annotations__",), updated=())
    def wrapper(*args, **kwargs):
        ctx = _get_current_ctx()
        runnable: Runnable = f(*args, **kwargs)
        result = None
        match ctx:
            case QPU():
                result = runnable.run(qpu=ctx)
            case Circuit():
                ctx.gates.append(runnable)
            case _:
                raise TypeError
        return result

    wrapper.__name__ = name
    wrapper.__qualname__ = name

    return wrapper


had = export(_create_ctx_function(Hadamard, "had"))
phase_shift = export(_create_ctx_function(PhaseShift, "phase_shift"))
measure = export(_create_ctx_function(Measure, "measure"))



from abc import ABC, abstractmethod

from quantalpy.qpu import QPU
import quantalpy.typing as qpt


class Runnable(ABC):
    @property
    def ends_with_measure(self) -> bool:
        return False

    @abstractmethod
    def run(self, qpu: QPU) -> qpt.MeasureOutcome | None:
        pass



import typing as t

from quantalpy.qpu import QPU
from quantalpy.runnable import Runnable
import quantalpy.typing as qpt
from quantalpy.utils import export


@export
class Measure(Runnable):
    def __init__(self, index: int | slice | t.Iterable[int]) -> None:
        self.index = index

    @property
    def ends_with_measure(self) -> bool:
        return True

    def run(self, qpu: QPU) -> qpt.MeasureOutcome:
        match self.index:
            case int():
                return qpu.measure(index=self.index)
            case slice():
                return tuple(
                    qpu.measure(index=i)
                    for i in range(*self.index.indices(qpu.n_qubits))
                )
            case _:
                return tuple(qpu.measure(i) for i in self.index)



from collections.abc import Collection

import numpy as np

from quantalpy.circuit import Circuit
from quantalpy.unary_qubit_gate import Hadamard, PhaseShift
from quantalpy.binary_qubit_gate import ControlledQubitGate
from quantalpy.utils import export


@export
def qft(indices: Collection[int]) -> Circuit:
    c = Circuit()

    for i in indices:
        c.gates.append(Hadamard(i))
        for a, x in enumerate(range(2, len(indices) + 1 - i), start=1):
            gate = PhaseShift(phi=2 * np.pi / (2**x), index=i)
            c.gates.append(ControlledQubitGate(gate=gate, control=i + a))

    return c



from collections import defaultdict
from quantalpy.runnable import Runnable
from quantalpy.qpu import QPU
from quantalpy.ctx import ContextMixin
from quantalpy.utils import export

import quantalpy.typing as qpt


@export
class Circuit(ContextMixin, Runnable):
    def __init__(self, gates: list[Runnable] | None = None) -> None:
        self.gates = gates or []

    @property
    def ends_with_measure(self) -> bool:
        if not self.gates:
            return False
        return self.gates[-1].ends_with_measure

    def run(self, qpu: QPU) -> qpt.MeasureOutcome | None:
        result = None
        for gate in self.gates:
            result = gate.run(qpu=qpu)
        return result

    def run_multiple(
        self, qpu: QPU, n: int, normalize: bool = True
    ) -> dict[qpt.MeasureOutcome, int]:
        if not self.ends_with_measure:
            raise RuntimeError(
                "Method can only be called if the circuit ends with a measurement"
            )
        results = defaultdict(int)
        for i in range(n):
            qpu.reset()
            results[self.run(qpu)] += 1
        qpu.reset()
        if normalize:
            return {k: v / n for k, v in results.items()}
        return dict(results)



import sys


def export(fn):
    mod = sys.modules[fn.__module__]
    if hasattr(mod, "__all__"):
        mod.__all__.append(fn.__name__)
    else:
        mod.__all__ = [fn.__name__]
    return fn


export(export)



import numpy as np
import numpy.typing as npt

from quantalpy.runnable import Runnable
from quantalpy.qpu import QPU
from quantalpy.utils import export


@export
class UnaryQubitGate(Runnable):
    def __init__(self, matrix: npt.ArrayLike, index: int) -> None:
        self.matrix = np.asarray(matrix, dtype=complex)
        self.index = index

    def run(self, qpu: QPU) -> None:
        qpu.apply_unary_qubit_operator(index=self.index, matrix=self.matrix)


@export
class Hadamard(UnaryQubitGate):
    def __init__(self, index: int) -> None:
        had = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        super().__init__(matrix=had, index=index)


@export
class PhaseShift(UnaryQubitGate):
    def __init__(self, phi: float, index: int) -> None:
        had = np.array([[1, 0], [0, np.exp(1j * phi)]])
        super().__init__(matrix=had, index=index)
